/***************
adevs, a library supporting discrete event simulation
Copyright (C) 2001 James Nutaro

This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

Bugs, comments, and questions can be sent to nutaro@ece.arizona.edu
***************/
/**
A quantized integrator.  The integrator operates on a
double valued state and input value.  This integrator
is based on the DEVS integrator presented in 
"Theory of Quantized Systems: DEVS simulation of perceiving
agents", Cybernetics and Systems, Volume 31, Number 6, Sept 2000.

<p>The integration function treats values less than EPSILON to be 
equal to zero for the purposes of computing internal transitions.
In particular, an integrator whose ta() < EPSILON is treated as
having ta() = 0.0.  
*/

class qintg {

public:
/// Constructor.  Creates an uninitialized integrator.
	inline qintg () { setParams(0.1); }
/**
Initialize the integrator with state q0 and input value dq0.
*/
	void init (double q0, double dq0 = 0.0);
/**
Get the time until the next quantum change in state, assuming constant
input.
*/
	inline double ta () const { return sigma; }
/**
Evaluate the integral over dt units of time with input dq.
dt should be less than ta(). 
*/
	void integ (double dt, double dq);
/**
Get the output value ta() units of time after the last state change
(at the next internal event time, tL + ta()).
*/
	inline double out() const { return (n+sign(x))*D; }
/// Get the output value for time tL + dt
	inline double out(double dt) const { 
		if (dt < sigma) return qstate();
		return (n+sign(x))*D; 
		}
/// Get the current state of the integrator
	inline double state() const { return q; }
/// Get the current quantized state of the integrator
	inline double qstate() const { return n*D; }
/**
Set the integrator parameters.  Arguments are quantum size D (default is 0.1) and
hysteresis e (if < 0, defaults to D / 2.0).
*/
	void setParams (double D, double e = -1.0);

private:
	double q, x, D, n, sigma, hyst;

	void compute_ta ();
	static double sign(double y);

};
