/***************
adevs, a library supporting discrete event simulation
Copyright (C) 1999, 2000, 2001 James Nutaro

This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

Bugs, comments, and questions can be sent to nutaro@ece.arizona.edu
***************/
/**
The devs class provides some fundamental behavior needed by its
network and atomic sub-classes.  Generally, none of the methods
here should be overridden.
*/

class devs: public object {

public:		
/// Enumeration of model types
		enum ModelType { UNKNOWN, ATOMIC, NETWORK };
/// Constructor.  Default model type is UNKNOWN
		devs (ModelType type = UNKNOWN);
/// Constructor.  Assigns the given name to the model.
		devs (const char* name, ModelType type = UNKNOWN);
/// Sets the model's name.
		inline void setName (const char* name) { this->name = name; }
/// Returns time elapsed since the last event.  e = t - tL
		inline stime_t elapsed () const { return t - tL; }
/// Returns time remaining until the next event. s = tN - t
		inline stime_t sigma () const { return tN - t; }
/// Returns the time allowed to remain in the current state. ta = tN - tL
		inline stime_t ta () const { return tN - tL; }
/// Returns the next event time.
		inline stime_t timeNext () const { return tN; }
/// Returns the last event time.
		inline stime_t timeLast () const { return tL; }
/// Returns the current simulation time. 
		inline stime_t timeCurrent () const { return t; }
/// Reschedules the model such that tN = t + dt
		void hold (stime_t dt);
/// Sets tN = INFINITY.
		inline void passivate () { tN = INFINITY; }
/**
Returns the input available on the port.
The bag should not be deleted, it is still owned the devs.
*/
		bag* inputOnPort (port_t inport) const;
/**
Writes a value to the output port.  If gc is false, object will not
be garbage collected.  Default is to garbage collect object.
*/
		void output (port_t outport, object* value, bool gc = true);
/**
Writes a clone() of the output value to the port.  Fails
with a CloneException if the object does not support the clone()
method.
*/
		void output (port_t outport, const object& value) {
		   output (outport, value.clone ());
		   }
/// Add an input port
		virtual void addInputPort (port_t inport);
/// Add an output port
		virtual void addOutputPort (port_t outport);
/// Returns true if the port is found, false otherwise
		bool hasInputPort (port_t inport) const;
/// Returns true if the port is found, false otherwise
		bool hasOutputPort (port_t outport) const;
/// Get the name of the model.
		inline const char* getName () const { return name.c_str (); }
/// Indicate that this model should be assigned its own thread of execution
		void setThreaded (bool value) { isThreaded = value; }
/// Returns true if this model will be assigned its own thread of execution
		bool getThreaded () const { return isThreaded; }
/// Method to support the simulator.
		const port_t* outputPorts (int *count);
/// Method to support the simulator.
		bag* outputOnPort (port_t outport) const;
/// Method to support the simulator.
		void clearPorts ();
/// Method to support the simulator.
		void clearPorts (indexed& garbage);
/// Method to support the simulator
		void emptyGarbage ();
/// Method to support the simulator.
		inline void setTime (stime_t t) { this->t = t; }
/// Method to support the simulator.
		inline void setTimeLast (stime_t tL) { this->tL = tL; }
/// Method to support the simulator.
		inline void setTimeNext (stime_t tN) { this->tN = tN; }
/**
Method to support the simulator.  Indicates that the model should reset its state.  
t indicates the initial time for the model.
*/
		virtual void _reset (stime_t t0 = 0.0);
/// Method to support the simulator.
		void input (port_t inport, object* value);
/// Method to support the simulator.
		virtual ~devs ();
/// Get the typecode for this model.
		ModelType getType () const { return typeCode; }

private:	
		cstr name;	
		set inports;
		set outports;
		stime_t tL;
		stime_t tN;
		stime_t t;
		bool adjust_outports;
		port_t* outport_tabs;
		ModelType typeCode;
		indexed garbage;
		bool isThreaded;

};
