/***************
adevs, a library supporting discrete event simulation
Copyright (C) 1999, 2000, 2001 James Nutaro

This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

Bugs, comments, and questions can be sent to nutaro@ece.arizona.edu
***************/
/**
Base class for coordinators and simulators.  Used internally.
Simulators are expected to be single threaded, except where noted.
*/
class simulator: public object {

public:	
/// Types are atomic simulator (SIM), coordinator (COORD), and their parallel variants (PSIM, PCOORD)
		enum SimType { UNKNOWN, SIM, COORD, PSIM, PCOORD };
/// Constructor.  Default type is UNKNOWN. 
		simulator (SimType type = UNKNOWN);
/**
Compute the simulator's output for time t.  Does not change the model state.
The simulator may fork a thread to do the I/O computation.
*/
		virtual void computeNextOutput (stime_t t) = 0;
/**
Compute an output prior to the next state change.  Assumes the next call will be
computeTransition (t).  Simulator may fork a thread to do the I/O computation.
*/
		virtual void computeIO (stime_t t) = 0;
/**
Compute the next state (assumes computeIO was called just prior).
The simulator may fork a thread to do the state change computation.
*/
		virtual void computeTransition (stime_t t) = 0;
/// Reset the simulator based on the model's current state
		virtual void reset (stime_t t0 = 0.0) = 0;
/// Add an input to the inport bag.
		virtual void input (port_t inport, object* value);
/**
Cleanup I/O related garbage at the end of cycle. 
*/
		virtual void endOfCycle (indexed& garbage);
/// Get the output on the assocated model port
		virtual bag* outputOnPort (port_t outport);
/// Get a list of the model outputs.  Returns an array of port_t types with count elements.
		virtual const port_t* outputPorts (int* count);
/// Get the attached model's next event time
		virtual stime_t timeNext ();
/// Get the attached model's sigma
		virtual stime_t sigma ();
/// Get the attached model's time of last event
		virtual stime_t timeLast ();
/// Get the attached model.
		virtual devs* getModel ();
/// If this is a multicomponent simulator, clear the simulator's schedule
		virtual void clear_schedule () { }
/// Get the simulator's type code
		inline SimType getType () const { return typeCode; }
/// Delete the simulator (but not its attached model)
		virtual ~simulator ();
/// Join with any forked threads (see computeIO, computeNextOutput, and computeTransition)
		virtual void wait () { }
protected:	
/// Attached a model to this simulator
		void setModel (devs* model);
		
private:	devs* model;
		SimType typeCode;
};
