/***************
adevs, a library supporting discrete event simulation
Copyright (C) 1999, 2000, 2001 James Nutaro

This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

Bugs, comments, and questions can be sent to nutaro@ece.arizona.edu
***************/
#include <adevs/adevs.h>
using namespace adevs;

/**
It is important to avoid calls to virtual DEVScoordinator methods
within the DEVScoordinator object.  This may result in deadlocks
during parallel simulation due to multiple attempts to acquire
the monitor for the parallel object.
*/

void DEVScoordinator::computeNextOutput (stime_t t) {

simulator* sim;
int count;
const port_t* ports;

// get everything that is imminent, and compute its next output
// the simulator might fork a thread to do the IO computation
while (sched.getSize () > 0 && sched.minPriority () == t) {
	sim = (simulator*)sched.removeMinimum ();
	sim->computeNextOutput (t);
	imm.add (sim);
	}

// Wait for parallel computation to end
for (unsigned i = 0; i < imm.getSize(); i++) {
	sim = (simulator*)imm[i];
	sim->wait ();
	}

// route the output of imminent models.  Size of the imminent
// list may change during route, so get its size before starting.
// All receivers are added to back of the imminent list, so
// this works fine.
unsigned size = imm.getSize ();
for (unsigned i = 0; i < size; i++) {
	sim = (simulator*)imm[i];
	ports = sim->outputPorts (&count);
	for (int j = 0; j < count; j++) 
		route (sim, ports[j]);
	}

// end of the I/O phase, now reschedule
// No need to reschedule the composite, since
// none of the models have changed state
for (unsigned i = 0; i < imm.getSize (); i++) {
	sim = (simulator*)imm[i];
	// reschedule the model
	schedule (sim);
	}

// end of method computeNextOutput

}

/**
It is important to realize that computeIO and
computeTransition methods are NOT independent.  A
computeTransition(t) call should ALWAYS be preceded
by a computeIO(t) call.  
*/

void DEVScoordinator::computeIO (stime_t t) {

simulator* sim;
int count;
const port_t* ports;

// make sure that t is not too far into the future
if (t > sched.minPriority () && sched.getSize() != 0) {
	ScheduleException err (model);
	throw err;
	}

// get everything that is imminent and compute its I/O function
// the simulator might fork a thread to do the IO computation
while (sched.getSize () > 0 && sched.minPriority () == t) {
	sim = (simulator*)sched.removeMinimum ();
	sim->computeIO (t);
	imm.add (sim);
	}

// Join with any forked threads
for (unsigned i = 0; i < imm.getSize(); i++) {
	sim = (simulator*)imm[i];
	sim->wait ();
	}
// route the output of imminent models, may add other imminents
// to the imminent set, to get the size now.  New imminent models are
// appended to the end of the set, so this will work fine.
unsigned size = imm.getSize ();
for (unsigned i = 0; i < size; i++) {
	sim = (simulator*)imm[i];
	ports = sim->outputPorts (&count);
	for (int j = 0; j < count; j++) 
		route (sim, ports[j]);
	}
		 
// end of the I/O phase 
}

void DEVScoordinator::computeTransition (stime_t t) {

simulator* sim;

// update the model time
model->setTime (t);

// If there are no imminents, just return. 
if (imm.getSize () == 0)
	return;

// compute the transition function for all of imminent and 
// externally activated models
// the simulator might fork a thread to do the transition computation
for (unsigned i = 0; i < imm.getSize (); i++) {
	sim = (simulator*)imm[i];
	sim->computeTransition (t);
	}
/// Join with threads and reschedule the models that changed state
for (unsigned i = 0; i < imm.getSize (); i++) {
	sim = (simulator*)imm[i];
	sim->wait();
	schedule (sim);
	}
// Make changes to the model structure
makeStructureChanges (t);
// update the models tL
model->setTimeLast (t);
// set the resulting tN for the network model.
scheduleComposite ();

// done computing the transition function!
}

// Updates the simulation schedule to reflect changes in the network
// structure.
void DEVScoordinator::makeStructureChanges (stime_t t) {

// Get the models to add and remove
model->makeStructureChanges (new_models, del_models);

// For every model that is to be removed
while (del_models.getSize () > 0) {
	devs* d = (devs*)del_models.remove (0);
	// remove the model from the schedule
	simulator* sim = removeModelFromSchedule (d);
	if (sim != NULL) {
		// if the simulator was imminent, remove it from the imm set
		if (imm.instanceCount (*sim) != 0) 
			imm.remove (imm.indexOf (*sim));
		// Return the simulator to the pool
		sim_pool.unwrap (sim);
		}
	// delete the model (its ports will be cleaned up at endOfCycle() or next reset())
	garbage_models.add (d);
	}
// For every model that is to be added
while (new_models.getSize () > 0) {
	devs* d = (devs*)new_models.remove (0);
	// put the model in its initial state
	d->_reset (t);
	// if the model is not passive, schedule it
	if (d->timeNext () != INFINITY) {
		simulator* sim = sim_pool.wrap (d);
		sim->reset (t);
		schedule (sim);
		}
	}

// done making the structure changes!
}

simulator* DEVScoordinator::removeModelFromSchedule (devs* d) {
	
// find the simulator for this model
simulator* sim = sim_pool.find (*d);
// if the model has a simulator
if (sim != NULL) {
	if (sched.isEnqueued (*sim)) {
		// remove the simulator from the schedule
		sched.adjust (*sim, pqueue::min_priority);
		sched.removeMinimum ();
		}
	// if this simulator maintains a component schedule, clear the simulator's schedule.
	sim->clear_schedule ();
	}
// Return the simulator attached to this model
return sim;

}

// Descend the simulation tree and clear all of the schedules.
// Providing an internal clear schedule method allows decendents
// to create a threadsafe version of the external clear_schedule
// method
void DEVScoordinator::clear_schedule () {

while (sched.getSize () > 0) {
	simulator* sim = (simulator*)sched.removeMinimum ();
	sim->clear_schedule ();
	sim_pool.unwrap (sim);
	}
// No simulators left, so clear the imminent set.
imm.removeAll ();

}

void DEVScoordinator::reset (stime_t t0) {

const set* models = model->getComponents ();

// clear the schedule, free up the simulators attached to scheduled
// models.
clear_schedule ();

// reset and then reschedule the component models that are not passive.
for (unsigned i = 0; i < models->getSize (); i++) {
	devs* d = (devs*)models->getObj (i);
	if (d->timeNext () < INFINITY) {
		simulator* sim = sim_pool.wrap (d); 
		sim->reset (t0);
		schedule (sim);
		}
	}

// set tN for the network model.
scheduleComposite ();

// the simulator should be ready to go now
}

void DEVScoordinator::setModel (devsn* model) {

simulator::setModel (model);
this->model = (devsn*)model;

}

void DEVScoordinator::input (port_t inport, object* value) {

// this is a little funny.  Makes sure that any house keeping done
// by the simulator class gets executed.
simulator::input (inport, value);
// send the incoming event to the appropriate component models
route (model, port_value (inport, value));

}

void DEVScoordinator::schedule (simulator* sim) {

if (sim->timeNext () < model->timeCurrent ()) {
	ScheduleException err (sim->getModel ());
	throw err;
	}

// remove a passivated model from the queue
bool enqueued = sched.isEnqueued (*sim); 
if (sim->timeNext () == INFINITY && enqueued) {
	sched.adjust (*sim, pqueue::min_priority);
	sched.removeMinimum ();
	}
// otherwise, adjust the model's queue position or do an insert
else if (enqueued)
	sched.adjust (*sim, sim->timeNext ());
else if (sim->timeNext () < INFINITY)
	sched.insert (sim, sim->timeNext ());

}	

/**
Method to route a single event.
*/
void DEVScoordinator::route (devs* src_model, const port_value& pv) {

const indexed* n = model->computeNeighbors (*src_model, pv);
// if no neighbors, don't go on 
if (n != NULL) {
	// otherwise, route the messages
	for (unsigned i = 0; i < n->getSize (); i++) {
		EventReceiver* er = (EventReceiver*)n->getObj (i);
		if (er->getModel () == src_model) {
			FeedbackException err (src_model);
			throw err;
			} 
		// internal traffick or traffick arriving on an external input port
		else if (er->getModel () != model) {
			simulator* recv = sim_pool.find (*er->getModel ());
  			if (recv == NULL) recv = sim_pool.wrap (er->getModel ());
  			recv->input (er->getPort(), pv.value);
  			imm.add (recv);
			}
		// traffick that will appear on an external output port
		else 
			model->output (er->getPort(), pv.value, false);
		}
	}
}

/**
Method to route all of the events on a port.
*/
void DEVScoordinator::route (simulator* src, port_t port) {

bag* values;
devs* src_model;

// external event arriving on an external port
if (src == this) {
	src_model = model;
	values = model->inputOnPort (port);
	}
// internal event generated by a component
else {
	src_model = src->getModel ();
	values = src->outputOnPort (port);
	}

// route the messages
for (unsigned k = 0; k < values->getSize(); k++) {
	port_value pv (port, values->getObj(k));
	// compute the set of receiving ports
	route (src_model, pv);
	}

}

void DEVScoordinator::endOfCycle (indexed& garbage) {

// Do cleanup for this simulator
simulator::endOfCycle (garbage);
// Do end of cycle cleanup for all active component simulators
// frees up the model's simulator if the model has passivated.
for (unsigned i = 0; i < imm.getSize(); i++) {
	simulator* sim = (simulator*)imm[i];
	sim->endOfCycle (garbage);
	if (sim->timeNext () == INFINITY)
		sim_pool.unwrap (sim);
	}
// Clear the imminent set
imm.removeAll ();
// Delete all models removed during the structure change
garbage.append (garbage_models);
for (unsigned i = 0; i < garbage_models.getSize(); i++) {
	devs* d = (devs*)garbage_models[i];
	d->clearPorts (garbage);
	}
garbage_models.removeAll ();

}

DEVScoordinator::~DEVScoordinator () {

while (sched.getSize () > 0)
	sim_pool.unwrap ((simulator*)sched.removeMinimum ());
garbage_models.empty ();

}

void DEVScoordinator::scheduleComposite () {

if (sched.getSize () == 0) {
	model->passivate ();
	return;
	}

// schedule the network model based on the smallest component tN value

model->setTimeNext (sched.minPriority ());

}
