/***************
adevs, a library supporting discrete event simulation
Copyright (C) 1999, 2000, 2001 James Nutaro

This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

Bugs, comments, and questions can be sent to nutaro@ece.arizona.edu
***************/
#include <adevs/adevs.h>
using namespace adevs;

devssim::devssim (devs* model, stime_t t0):object () {

sim = NULL;
this->model = model;
reset (t0, true);

}

void devssim::computeNextOutput (indexed& output) {

int port_count;
const port_t* ports;

// compute the output at tN
sim->computeNextOutput (sim->timeNext ());
ports = sim->outputPorts (&port_count);

// Iterate throught the output ports, removing values
// and adding them to the list of results.  
for (int i = 0; i < port_count; i++) {
	bag* values = sim->outputOnPort (ports[i]);
	for (unsigned j = 0; j < values->getSize (); j++) {
		object* o = values->getObj(j)->clone ();
		output.add (new port_value (ports[i], o));
		}
	values->removeAll ();
	}

// Collect garbage but do not clean up.  Some models may require
// that these objects survive for the next state transition or output
// computation.
sim->endOfCycle (garbage);

// done computing the output at tN
}

void devssim::step (stime_t dt, indexed& input) {

// Note that everything passed to the step method is consumed.
// This should help to prevent tampering with the input
// value after they have been delivered to the simulator.
// The effects of the tampering will be largely 
// unpredictable, and probable very bad.

// Apply the inputs to the model.
for (unsigned i = 0; i < input.getSize (); i++) {
	port_value* in = (port_value*)input[i];
	sim->input (in->port, in->value);
	}
// Clear the input list
input.empty ();

// Step ahead to the desired time and compute the results.
step (dt);

}

void devssim::step (stime_t dt) {

// the root coordinator with stepwise advancement.
// advancement beyond the simulator's time of next
// event will have unpredictable effects on the simulator.

// get the next event time
t = t + dt;
if (dt < 0 || t > sim->timeNext ()) {
	ScheduleException schderr (model);
	throw schderr;
	}
compute_cycle (t);

} 

void devssim::run (stime_t t_max, unsigned cycles_max) {

// the root coordinator
unsigned cycle_count = 0;
// while more to compute
while (!halt () && t < t_max && cycle_count <= cycles_max) {
	cycle_count++;
	// get the next event time
	t = sim->timeNext ();
	if (t < sim->timeLast ()) {
		ScheduleException schderr (model);
		throw schderr;
		}
	compute_cycle (t);
	}
 
}

void devssim::reset (bool reset_model) {

reset (0.0, reset_model);

}

void devssim::reset (stime_t t0, bool reset_model) {

// delete the old simulator if one exists
if (sim != NULL) delete sim;

// create a new simulator
if (model->getType() == devs::ATOMIC) {
	DEVSsimulator* asim = new DEVSsimulator ();
	asim->setModel ((atomic*)model);
	sim = asim;
	}
else {
	DEVScoordinator* csim = new DEVScoordinator ();
	csim->setModel ((devsn*)model);
	sim = csim;
	}

// reset the model if necessary
if (reset_model)
	model->_reset (t0);

// reset the simulator
sim->reset (t0);
t = sim->timeLast ();

}

bool devssim::halt () {

if (isPassive ())
	return true;

return false;

}

stime_t devssim::timeCurrent () {

return t;

}

stime_t devssim::timeNext () {

return sim->timeNext ();

}

stime_t devssim::timeLast () {

return sim->timeLast ();

}

bool devssim::isPassive () {

return sim->timeNext () >= INFINITY;

}

devssim::~devssim () {

delete sim;
delete model;

}

void devssim::compute_cycle (stime_t t) {

// compute the I/O function
sim->computeIO (t);
// compute the transition function
sim->computeTransition (t);
// clean up at end of cycle
sim->endOfCycle (garbage);
garbage.empty ();

}

