/***********
Copyright (C) 1999, 2000  James Nutaro

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

Bugs, comments, and questions can be sent to nutaro@ece.arizona.edu
***************/
#include <adevs/adevs.h>
#include <time.h>
#include <stdlib.h>
#include <math.h>
#include "point.h"
#include "viewer.h"
#include "cell.h"
#include "cellspace.h"
#include <assert.h>
#include <iostream>
using namespace adevs;

#define NSIZE 8

const port_t cellspace::birthed = 1;
const port_t cellspace::isolated = 0;

cellspace::cellspace (int width, int height):
networkExec () {

addInputPort (birthed);
addInputPort (isolated);
this->width = width;
this->height = height;

}

/**
Reset/initialization function.
*/
void cellspace::reset () {

// Remove everything from the cellspace
object** keys = cells.getKeys ();
unsigned size = cells.keyCount ();
for (unsigned i = 0; i < size; i++) {
   cell* c = (cell*)cells.find (*(keys[i]));
   remove_cell (c);
   }
if (keys != NULL) delete [] keys;
assert (getComponents()->getSize () == 1);

// populate the cellspace
srand (clock ());
for (int x = 0; x < width; x += rand () % (width / 2)) {
   for (int y = 0; y < height; y += rand () % (height / 2)) {
      point p = map_point (x, y);
      createBlock (p);
      }
   }

passivate ();

}

void cellspace::delta_int () {

passivate ();

}

void cellspace::delta_conf () {

delta_ext ();

}

void cellspace::delta_ext () {

// remove the isolated cells
bag* values = inputOnPort (isolated);
for (unsigned i = 0; i < values->getSize(); i++) {
   point* p = (point*)values->getObj (i);
   remove_cell ((cell*)cells.find (*p));
   }

// add the cells who have living neighbors
hashtable new_cells;
indexed new_cell_list;
point l [NSIZE];
values = inputOnPort (birthed);
// for every cell that was birthed
for (unsigned i = 0; i < values->getSize(); i++) {
   // get its location
   point* p = (point*)values->getObj (i);
   // compute its neighbors
   getAdj (*p, l);
   // for each neighbor
   for (unsigned j = 0; j < NSIZE; j++) {
      // if this neighbor is not already in the cellspace  
      if (cells.find (l[j]) == NULL) {
         // if this neighbor is not already in the list of new cells
         if (new_cells.find (l[j]) == NULL) {
            // add the cell to the new cell list 
            cell* n = new cell (l[j], 0);
            new_cells.insert (&(n->getLocation()), n);
            new_cell_list.add (n);
            }
         // increment the living neighbor count for the cell 
         cell* n = (cell*)new_cells.find (l[j]);
         n->incrAlive ();
         }
      }
   }

// add all of the new cells to the cellspace
for (unsigned i = 0; i < new_cell_list.getSize(); i++) 
   add_cell ((cell*)new_cell_list[i]);

passivate ();

}

void cellspace::output_func () { }

/**
computeNeighbor function
*/
const indexed* cellspace::computeNeighbors (devs& src, const port_value& pv) {

// clear the neighbor list (we keep reusing it).
n.empty ();

// If a cell was isolated, send that output to this model
if (pv.port == cell::q)
   n.add (new EventReceiver (this, isolated));
// If a cell was birthed, send that output to this model
else if (pv.port == cell::b)
   n.add (new EventReceiver (this, birthed));
// otherwise, report the +/-1 neighbor to all of the src's neighbors
else {
   // cast the source of the event to a cell and get its location.
   cell* c = (cell*)&src;
   // compute the Moore neightborhood, adding an EventReceiver (model, port)
   // pair to the neightbor list for each adjacent cell.  If a non-existing cell is
   // to going to be the recipient of an event, that cell will be created
   // and added to the simulator with the appropriate neighbor count during
   // the external transition.
   point l [NSIZE];
   getAdj (c->getLocation(), l);

   for (int i = 0; i < NSIZE; i++) {
      cell* r = (cell*)cells.find (l[i]);
      if (r != NULL) 
         n.add (new EventReceiver (r, cell::n));
      }
   }
   
// return the list of EventReceivers to the simulator.
return &n;

}

/**
Create a cell and add it to the cellspace.
*/
cell* cellspace::add_cell (const point& loc, bool alive) {

// create the new cell.
cell* c = new cell (loc, alive);
// add the cell to the cellspace's set of components.
networkExec::add (c);
cells.insert (&(c->getLocation ()), c);
// return a pointer to the new cell.
return c;

}

/**
Add a cell to the cellspace
*/
void cellspace::add_cell (cell* c) {

// add the cell to the cellspace's set of components.
networkExec::add (c);
cells.insert (&(c->getLocation ()), c);

}

/**
Removes cell from the cellspace.
*/
void cellspace::remove_cell (cell* model) {

cells.remove (((cell*)model)->getLocation());
networkExec::remove (model);

}

/**
Fills the array of NSIZE points with the neighbors of p
*/
void cellspace::getAdj (const point& p, point* l) {

point n;
int offset = 0;
for (int i = -1; i < 2; i++) {
   n.x = p.x + i; n.y = p.y - 1;
   l[offset] = map_point (n.x, n.y);
   offset++;
   }
for (int i = -1; i < 2; i++) {
   n.x = p.x + i; n.y = p.y + 1;
   l[offset] = map_point (n.x, n.y);
   offset++;
   }
n.x = p.x + 1; n.y = p.y;
l[offset] = map_point (n.x, n.y);
offset++;
n.x = p.x - 1; n.y = p.y;
l[offset] = map_point (n.x, n.y);

}

/**
Creates a cross shaped block of cells around the indicated point.
*/
void cellspace::createBlock (const point& center) {

point l [NSIZE];
getAdj (center, l);
for (int i = 0; i < NSIZE; i++) {
   if (cells.find (l[i]) == NULL)
      add_cell (l[i], true);
   }

}

/**
Maps the point (x,y) into the cell's torrodial space.
*/
point cellspace::map_point (int x, int y) const {

point mapp;

int nx = x;
int ny = y;

if (x < 0) nx = width - 1;
else if (x >= width) nx = 0;

if (y < 0) ny = height - 1;
else if (y >= height) ny = 0;

mapp.set (nx, ny);
return mapp;

}

cellspace::~cellspace () {

object** keys = cells.getKeys ();
unsigned size = cells.keyCount ();

for (unsigned i = 0; i < size; i++) 
   cells.remove (*keys[i]);

delete[] keys;

n.empty ();

}
