/***********
Copyright (C) 1999, 2000  James Nutaro

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

Bugs, comments, and questions can be sent to nutaro@ece.arizona.edu
***************/
#include <string>
#include <iostream>
#include <stdlib.h>
#include <adevs/adevs.h>
#include "point.h"
#include "Exception.h"
#include "viewer.h"
#include "cell.h"
#include "cellspace.h"
using namespace adevs;

static const char* helpmsg = 
"Options: 	-h                  Print this message and exit
		-pause<seconds>     Pause for <seconds> between each game
		-step<seconds>      Pause for <seconds> between each generation
		-g<number>          Compute <number> of generations per game
		-G<height>x<width>  Use a <height> by <width> gameboard
		-s                  Run in screen saver mode.
		-root               Run in the root window.
		-thread<number>     Paralllel execution with <number> threads
Press 'q' to quit the game (won't work in the root window).";

// Number of generations to compute for each game.
static unsigned generations = 500;
// Pause time between games, measured in seconds.
static int pause_t = 5;
// Interval between simulation cycles,  measured in seconds.
static int step_t = 0;
// Threads for parallel simulation
static int threads = 0;

static void parseCommand (const char* cmd);

/**

The lifesim class extends the devssim class to incorporate support
for the dynamic cellspace, pauses between simulation cycles,
and simulation halts only after a given number of generations have been
observed.

*/
class lifesim: public devssim {

public:		lifesim (cellspace* world, unsigned generations, int step_time);

// Overrides the halt method from devssim.  Returns true after
// the indicated number of generations have been observed.
		bool halt ();
// Resets the simulator. 
		void reset ();
// Method to indicate if the user has pressed the quit key.
		bool hasQuit ();

		~lifesim ();

private:	
// the number of generations observed so far.
		unsigned count; 
// the total number of generations to observe
		unsigned limit;
// the time to pause between simulation cycles.
		int step_t;
// flag indicating whether the user has wants to quit.
		bool quit;

};

lifesim::lifesim (cellspace* world, unsigned generations, int step_time):
devssim (new DScoupled (world)) { 

count = 0;
limit = generations;
step_t = step_time;
quit = false;

}

/**
Reset/initialization function.
*/
void lifesim::reset () {

// Call the parent's reset method.  Mandatory.
devssim::reset ();

// Reset the generation counter.  Clear the quit flag.
count = 0;
quit = false;

}

/**
The halt method is called by the simulator at the end of
each cycle.  If halt () returns true, the run () method returns.
*/
bool lifesim::halt () {

if (cell::view->checkEvents ()) {
   quit = true;
   return true;
   }

if (step_t > 0)
   sleep (step_t);

return (devssim::halt () || count++ > limit);

}

bool lifesim::hasQuit () { return quit; }

lifesim::~lifesim () { }


/**

Main function.  Parses the command line for some game
specific options before creating the simulation and the
viewer.

*/

int main (int argc, char** argv) {

for (int i = 1; i < argc; i++) 
   parseCommand (argv[i]);

// Create the display window.  Exit on an error.
try { 
   cell::view = new viewer (argv, argc);
   cell::view->checkEvents ();
   }
catch (Exception& err) {
   std::cerr << err.getError () << std::endl;
   exit (-1);
   }

// Create the simulator.
lifesim sim (new cellspace (cell::view->getWidth (), 
		cell::view->getHeight ()), generations, step_t);

// Initialize and the run the simulator.  
sim.setThreadLimit (threads);
sim.reset ();
sim.run ();

// Continue to reset and run the simulator until the user asks
// us to quit.

while (!sim.hasQuit ()) {
   if (pause_t > 0)
      sleep (pause_t);
   cell::view->blank ();
   sim.reset ();
   sim.run ();
   }
// clean up and exit.
delete cell::view;


}

static void parseCommand (const char* cmd) {

int val;
unsigned i, j;
std::string arg = cmd;
char value [100];
size_t loc;

loc = arg.find ("-h", 0, strlen ("-h"));

if (loc != std::string::npos) {

   std::cout << helpmsg << std::endl;
   exit (1);

   }

loc = arg.find ("-g", 0, strlen ("-g"));

if (loc != std::string::npos) {

   val = 0;
   j = 0;

   for (i = strlen ("-g"); i < arg.length () && isdigit (cmd[i]); i++) 
      value[j++] = cmd[i];   

   value[j] = (char)NULL;

   if (strlen (value) > 0)
      val = atoi (value);

   if (val > 0) generations = (unsigned)val; 

   return;

   }

loc = arg.find ("-step", 0, strlen ("-step"));

if (loc != std::string::npos) {

   val = 0;
   j = 0;

   for (i = strlen ("-step"); i < arg.length () && isdigit (cmd[i]); i++) 
      value[j++] = cmd[i];   

   value[j] = (char)NULL;

   if (strlen (value) > 0)
      val = atoi (value);

   if (val >= 0) step_t = val; 

   return;

   }

loc = arg.find ("-pause", 0, strlen ("-pause"));

if (loc != std::string::npos) {

   val = 0;
   j = 0;

   for (i = strlen ("-pause"); i < arg.length () && isdigit (cmd[i]); i++) 
      value[j++] = cmd[i];   

   value[j] = (char)NULL;

   if (strlen (value) > 0)
      val = atoi (value);

   if (val >= 0) pause_t = val; 

   return;

   }

loc = arg.find ("-thread", 0, strlen ("-thread"));

if (loc != std::string::npos) {

   val = 0;
   j = 0;

   for (i = strlen ("-thread"); i < arg.length () && isdigit (cmd[i]); i++) 
      value[j++] = cmd[i];   

   value[j] = (char)NULL;

   if (strlen (value) > 0)
      val = atoi (value);

   if (val >= 0) threads = val; 

   return;

   }

}

