/***********
Copyright (C) 1999, 2000  James Nutaro

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

Bugs, comments, and questions can be sent to nutaro@ece.arizona.edu
***************/
#include "message.h"
#include "process.h"
#include <iostream>
#include <assert.h>
using namespace adevs;

const port_t process::send = 0;
const port_t process::recv = 1;
const port_t process::prog_in = 2;
const port_t process::prog_out = 3;

process::process (unsigned pid, unsigned num_procs)
:atomic ("process") {

this->pid = pid;
this->num_procs = num_procs;
sync = (bool*) malloc (sizeof(bool) * num_procs);
waiting = new queue * [num_procs];
for (unsigned i = 0; i < num_procs; i++) waiting[i] = new queue ();
addInputPort (recv);
addOutputPort (send);
addInputPort (prog_in);
addOutputPort (prog_out);

}

/**

Initial state:

sync[i] = false for all i in the range [0, num_procs]
length (waiting[i]) = 0 for all i in [0, num_procs]
length (safe) = 0
c = 0

Implementation details include clearing the I/O buffers
and setting tN = INFINITY

*/

void process::reset () {

c = 0;
safe.empty ();
pout.empty (); 
dout.empty ();
for (unsigned i = 0; i < num_procs; i++) {
   sync[i] = false;
   waiting[i]->empty ();
   }
passivate ();

}

/**

Internal transition used to support implementation only.
Clear the channel buffers and set tN = INFINITY.

*/
void process::delta_int () {

dout.removeAll ();
pout.removeAll ();
passivate ();

}

/**

External transition function produces 
a recv(message) event for all inputs

*/
void process::delta_ext () {

bag* values = inputOnPort (recv);
for (unsigned i = 0; i < values->getSize (); i++) {
   const message* m = (message*)values->getObj (i);
   _recv (*m);
   }

values = inputOnPort (prog_in);
for (unsigned i = 0; i < values->getSize (); i++) {
   const message* m = (message*)values->getObj (i);
   _recv (*m);
   }

}

void process::delta_conf () {

delta_int ();
delta_ext ();

}

/**
Output everything in the output buffers.
*/
void process::output_func () {

for (unsigned i = 0; i < pout.getSize (); i++)
   output (prog_out, pout[i]);
for (unsigned i = 0; i < dout.getSize (); i++)
   output (send, dout[i]);

}

process::~process () {

unsigned unprocessed = safe.getSize();
safe.empty (); dout.empty (); pout.empty ();
for (unsigned i = 0; i < num_procs; i++) {
   unprocessed = unprocessed + waiting[i]->getSize();
   waiting [i]->empty ();
   delete waiting[i];
   }
delete waiting;
free (sync);

std::cout << pid << " has " << unprocessed << " unprocessed messages " << std::endl;

}

/**
send(message) primitive to support the distribute system model 
that is being used.
*/
void process::_send (object* msg, port_t port) {

if (port == send) dout.add (msg);
else pout.add (msg);
hold (0);

}

/**
The meat of the algorithm!
*/

// on receiving a message m
void process::_recv (const message& msg) {

message* m = new message (msg);

// if no message has been receiving from the process or program with m.pid
if (!sync[m->pid]) {

   // if it from the attached program (e.g. an internal event)
   if (m->pid == pid) {
      m->c = c;
      c++;
      _send (new message (*m), send);
      }
   // otherwise if it is from another process (e.g. a recv event)
   else {
      if (c < m->c) c = m->c;
      c++;
      // if we haven't already sent a message, send a control message to
      // avoid a possible deadlock
      if (!sync[pid]) {
         sync[pid] = true;
         message* control = new message ();
         control->pid = pid; control->control = true;
         control->c = c;
         c++;
         _send (control, send);
         }
      }
   // set the sync flag to reflect receiving from this process
   sync[m->pid] = true;
   // add the message to the list of events that it is safe to process,
   // don't process control messages
   if (!m->control)
      safe.add (m);
   // try to execute the messages that have accumulated so far
   execute ();
   }

// otherwise just add the message to the appropriate waiting queue
else if (!m->control) waiting[m->pid]->add (m);

}

// to execute a message
void process::execute () {

// if all the sync flags are true proceed, otherwise just return
bool syncd = true;
for (unsigned i = 0; i < num_procs && syncd; i++)
   syncd = syncd && sync[i];

if (!syncd) return;

// create a queue to hold the sorted safe list
queue* sorted = new queue ();

// sort the safe list (the list should very short for these examples, 
// so just use a bubble sort)
while (safe.getSize () > 0) {
   unsigned min = 0;
   for (unsigned k = 0; k < safe.getSize (); k++) {
      message* m = (message*)safe[k];
      if (*m < (*(message*)safe[min])) min = k;
      }
   sorted->add (safe.remove (min));
   }

// send the sorted list to the program for execution
_send (sorted, prog_out);

// clear all of the sync flags
for (unsigned i = 0; i < num_procs; i++) sync[i] = false;

// for each waiting queue, get the first message of the list (if length (waiting[i]) > 0)  
// and generate a recv event.  This will generate outgoing queues until the synchronization
// conditions are no longer met.
for (unsigned i = 0; i < num_procs; i++) {
   if (waiting[i]->getSize () > 0) {
      message* m = (message*)waiting[i]->remove ();
      _recv (*m);
      delete m;
      }
   }

}
